
from . import PhishingTrackerCertificate
from . import PhishingTrackerDig
from . import PhishingTrackerSmtp
from . import PhishingTrackerWeb
from . import PhishingTrackerWhois

import copy


class PhishingTrackerAnalyzers:

    def __init__(self, __logger=None):
        global logger
        logger = __logger

    def analyzers(self, data, analyzers=None):

        logger.info('analyzers: {}'.format(', '.join(analyzers)))

        analyzer_data = {}

        if 'whois' in analyzers:
            logger.debug('whois analyzer: {}'.format(data['meta']['domain_name']))
            analyzer_data['whois'] = PhishingTrackerWhois.analyzer(data['meta']['domain_name'])

        if 'dig' in analyzers or 'smtp' in analyzers:
            analyzer_data['dig'] = {}
            logger.debug('dig-domain analyzer: {}'.format(data['meta']['domain_name']))
            analyzer_data['dig']['domain_name'] = PhishingTrackerDig.analyzer(record=data['meta']['domain_name'])
            if data['meta']['host_name'] != data['meta']['domain_name']:
                logger.debug('dig-host analyzer: {}'.format(data['meta']['host_name']))
                analyzer_data['dig']['host_name'] = PhishingTrackerDig.analyzer(record=data['meta']['host_name'])

        if 'smtp' in analyzers and 'dig' in analyzer_data:
            analyzer_data['smtp'] = {}
            if 'domain_name' in analyzer_data['dig'] and len(analyzer_data['dig']['domain_name']['MX']) > 0:
                logger.debug('smtp-domain analyzer: {}'.format(', '.join(analyzer_data['dig']['domain_name']['MX']) ))
                analyzer_data['smtp']['domain_name'] = PhishingTrackerSmtp.analyzer(analyzer_data['dig']['domain_name']['MX'])
            if 'host_name' in analyzer_data['dig'] and len(analyzer_data['dig']['host_name']['MX']) > 0:
                logger.debug('smtp-host analyzer: {}'.format(', '.join(analyzer_data['dig']['host_name']['MX']) ))
                analyzer_data['smtp']['host_name'] = PhishingTrackerSmtp.analyzer(analyzer_data['dig']['host_name']['MX'])

        http_url = data['meta']['url_decode']
        if http_url.startswith('https://'):
            http_url = http_url.replace('https://', 'http://')

        https_url = data['meta']['url_decode']
        if https_url.startswith('http://'):
            https_url = https_url.replace('http://', 'https://')

        if 'http' in analyzers:
            logger.debug('http-get analyzer: {}'.format(http_url))
            analyzer_data['http'] = PhishingTrackerWeb.analyzer(url=http_url)

        if 'https' in analyzers or 'https_certificate' in analyzers:
            logger.debug('https-get analyzer: {}'.format(https_url))
            analyzer_data['https'] = PhishingTrackerWeb.analyzer(url=https_url)

        if 'https_certificate' in analyzers:
            logger.debug('https-certificate analyzer: {}'.format(https_url))
            analyzer_data['https_certificate'] = PhishingTrackerCertificate.analyzer(hostname=data['meta']['host_name'])

        return copy.copy(analyzer_data)

    def analyzers_report(self, data, analyzer_report_sets_expected=None, analyzer_report_sets_not_expected=None):

        status = []
        analyzers_data = data['analyzers']

        if 'dig' in analyzers_data:
            if 'host_name' in analyzers_data['dig']:
                if len(analyzers_data['dig']['host_name']['A']) > 0:
                    status.append('dns_hostname_a_record')
                if len(analyzers_data['dig']['host_name']['AAAA']) > 0:
                    status.append('dns_hostname_aaaa_record')
                if len(analyzers_data['dig']['host_name']['CNAME']) > 0:
                    status.append('dns_hostname_cname_record')
                if len(analyzers_data['dig']['host_name']['MX']) > 0:
                    status.append('dns_hostname_mx_record')
                if len(analyzers_data['dig']['host_name']['NS']) > 0:
                    status.append('dns_hostname_ns_record')
                if len(analyzers_data['dig']['host_name']['TXT']) > 0:
                    status.append('dns_hostname_txt_record')

            if 'domain_name' in analyzers_data['dig']:
                if len(analyzers_data['dig']['domain_name']['A']) > 0:
                    status.append('dns_domainname_a_record')
                if len(analyzers_data['dig']['domain_name']['AAAA']) > 0:
                    status.append('dns_domainname_aaaa_record')
                if len(analyzers_data['dig']['domain_name']['CNAME']) > 0:
                    status.append('dns_domainname_cname_record')
                if len(analyzers_data['dig']['domain_name']['MX']) > 0:
                    status.append('dns_domainname_mx_record')
                if len(analyzers_data['dig']['domain_name']['NS']) > 0:
                    status.append('dns_domainname_ns_record')
                if len(analyzers_data['dig']['domain_name']['TXT']) > 0:
                    status.append('dns_domainname_txt_record')

        if data['meta']['host_name'] == data['meta']['domain_name']:
            status.append('dns_hostname_eq_dns_domainname')

        if 'http' in analyzers_data:
            if 'exception' in analyzers_data['http']:
                status.append('http_exception')
            if 'response' in analyzers_data['http'] and 'status_code' in analyzers_data['http']['response']:
                status.append('http_hostname_{}_response'.format(analyzers_data['http']['response']['status_code']))

        if 'https' in analyzers_data:
            if 'exception' in analyzers_data['https']:
                status.append('https_exception')
            if 'response' in analyzers_data['https'] and 'status_code' in analyzers_data['https']['response']:
                status.append('https_hostname_{}_response'.format(analyzers_data['https']['response']['status_code']))

        if 'https_certificate' in analyzers_data:
            if 'exception' in analyzers_data['https_certificate']:
                status.append('https_certificate_exception')
            https_certificate_names = []
            if 'subject' in analyzers_data['https_certificate'] and 'commonName' in analyzers_data['https_certificate']['subject']:
                https_certificate_names.append(analyzers_data['https_certificate']['subject']['commonName'])
            if 'subjectAltName' in analyzers_data['https_certificate'] and 'DNS' in analyzers_data['https_certificate']['subjectAltName']:
                https_certificate_names.extend(analyzers_data['https_certificate']['subjectAltName']['DNS'])
            https_certificate_name_found = False
            for https_certificate_name in https_certificate_names:
                if https_certificate_name == '*.{}'.format(data['meta']['domain_name']) or https_certificate_name == data['meta']['host_name']:
                    https_certificate_name_found = True
                    break
            if https_certificate_name_found is False:
                status.append('https_certificate_hostname_mismatch')

        if 'smtp' in analyzers_data:
            if 'exception' in analyzers_data['smtp']:
                status.append('smtp_exception')
            if 'domain_name' in analyzers_data['smtp'] and analyzers_data['smtp']['domain_name'] is not None and len(analyzers_data['smtp']['domain_name']) > 0:
                for item_k, item_v in analyzers_data['smtp']['domain_name'].items():
                    if 'status_code' in item_v and item_v['status_code'] in [220]:
                        if 'smtp_domainname_active' not in status:
                            status.append('smtp_domainname_active')
            if 'host_name' in analyzers_data['smtp'] and analyzers_data['smtp']['host_name'] is not None and len(analyzers_data['smtp']['host_name']) > 0:
                for item_k, item_v in analyzers_data['smtp']['host_name'].items():
                    if 'status_code' in item_v and item_v['status_code'] in [220]:
                        if 'smtp_hostname_active' not in status:
                            status.append('smtp_hostname_active')

        if 'whois' in analyzers_data:
            if 'exception' in analyzers_data['whois']:
                status.append('whois_exception')
            else:
                status.append('whois_domainname_record')

        analyzer_report_sets = {
            'expected': None,
            'not_expected': None
        }

        if analyzer_report_sets_expected and type(analyzer_report_sets_expected) is list:
            analyzer_report_sets['expected'] = False
            for item_set in analyzer_report_sets_expected:
                if type(item_set) is str:
                    item_set = [item_set]
                item_set_count = 0
                for item in item_set:
                    if item in status:
                        item_set_count += 1
                if item_set_count == len(item_set):
                    analyzer_report_sets['expected'] = True
                    break

        if analyzer_report_sets_not_expected and type(analyzer_report_sets_not_expected) is list:
            analyzer_report_sets['not_expected'] = False
            for item_set in analyzer_report_sets_not_expected:
                if type(item_set) is str:
                    item_set = [item_set]
                item_set_count = 0
                for item in item_set:
                    if item in status:
                        item_set_count += 1
                if item_set_count == len(item_set):
                    analyzer_report_sets['not_expected'] = True
                    break

        report_data = {
            'reference': data['meta']['reference'],
            'reports': status,
            'analyzer_report_sets': analyzer_report_sets
        }
        return copy.copy(report_data)




import time
import logging

from . import NAME
from . import PhishingTrackerException


class ColoredLoggingFormatter(logging.Formatter):

    color_line = '\x1b[90m'  # grey
    color_reset = '\x1b[0m'  # reset

    def __init__(self, **kwargs):
        if 'fmt' in kwargs:
            kwargs['fmt'] = '{}{}{}'.format(self.color_line, kwargs['fmt'], self.color_reset)
        logging.Formatter.__init__(self, **kwargs)

    def format(self, record):

        levelname = record.levelname.upper()

        if levelname == 'CRITICAL':
            color_code = '\x1b[41m'  # white-on-red
        elif levelname == 'ERROR':
            color_code = '\x1b[31m'  # red
        elif levelname in ('WARNING', 'WARN'):
            color_code = '\x1b[33m'  # yellow
        elif levelname == 'INFO':
            color_code = '\x1b[36m'  # cyan
        elif levelname == 'DEBUG':
            color_code = '\x1b[37m'  # white
        else:
            color_code = '\x1b[90m'  # grey

        record.levelname = '{}{}{}'.format(color_code, levelname, self.color_line)

        return logging.Formatter.format(self, record)


class PhishingTrackerLogger:

    logger = None
    default_level = 'info'

    def __init__(self, level=None):

        if level is not None:
            log_level = level
        else:
            log_level = self.default_level

        logger_init = logging.getLogger(NAME)
        if logger_init.level > 0:
            self.logger = logger_init
            return

        logger_init.setLevel(logging.DEBUG)

        log_level = log_level.upper()
        stream_handler = logging.StreamHandler()

        if log_level == 'CRITICAL':
            stream_handler.setLevel(logging.CRITICAL)
        elif log_level == 'ERROR':
            stream_handler.setLevel(logging.ERROR)
        elif log_level in ('WARNING', 'WARN'):
            stream_handler.setLevel(logging.WARNING)
        elif log_level == 'INFO':
            stream_handler.setLevel(logging.INFO)
        elif log_level == 'DEBUG':
            stream_handler.setLevel(logging.DEBUG)
        elif log_level is not None:
            raise PhishingTrackerException('unknown loglevel value', log_level)
        else:
            stream_handler.setLevel(logging.NOTSET)

        formatter = ColoredLoggingFormatter(
            fmt='%(asctime)s - %(levelname)s - %(message)s',
            datefmt='%Y%m%dZ%H%M%S'
        )
        logging.Formatter.converter = time.gmtime

        stream_handler.setFormatter(formatter)
        logger_init.addHandler(stream_handler)

        self.logger = logger_init




import json
import datetime
import dateparser


def json_dumps(o, indent=None):

    def __json_dumps(o):
        if isinstance(o, datetime.datetime):
            return "{}-{}-{}".format(o.year, o.month, o.day)

    return json.dumps(o, default = __json_dumps, indent=indent)


def timestamp():
    return datetime.datetime.utcnow().strftime("%Y%m%dZ%H%M%S")


def datetime_parse(string):
    return dateparser.parse(string, settings={'TO_TIMEZONE': 'UTC'}).strftime("%Y-%m-%dT%H:%M:%S+00:00")




import random

# 2020-01-20
# https://techblog.willshouse.com/2012/01/03/most-common-user-agents/
raw_user_agents = """

Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.88 Safari/537.36
Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36
Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36
Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:71.0) Gecko/20100101 Firefox/71.0
Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.4 Safari/605.1.15

"""

def user_agent():
    lines = raw_user_agents.strip().split('\n')
    return random.choice(lines)




analyzers_available = ['dig', 'http', 'https', 'https_certificate', 'smtp', 'whois']

from .__name__ import NAME
from .__version__ import VERSION

from .exceptions import PhishingTrackerException
from .logger import PhishingTrackerLogger

from .file import PhishingTrackerFile
from .meta import PhishingTrackerMeta

from .certificate import PhishingTrackerCertificate
from .dig import PhishingTrackerDig
from .smtp import PhishingTrackerSmtp
from .web import PhishingTrackerWeb
from .whois import PhishingTrackerWhois

from .analyzers import PhishingTrackerAnalyzers
from .tests import PhishingTrackerTests

from .main import PhishingTracker




import re


class PhishingTrackerTests:

    def __init__(self, __logger=None):
        global logger
        logger = __logger

    def tests(self, data, tests=None):

        if tests is None or len(tests) == 0:
            return None

        logger.info('tests: {}'.format(', '.join(tests.keys())))
        test_data = {}

        for test_k, test_v in tests.items():
            if test_k not in data['analyzers']:
                logger.warn('skipping {} test, no related analyzer data available'.format(test_k))
                continue

            if test_k in tests and test_k not in test_data:
                logger.debug('testing {}'.format(test_k))
                response = self.__test(tests=test_v, data=data['analyzers'][test_k])
                if response:
                    test_data[test_k] = response

        return test_data

    def __test(self, tests, data):
        for test in tests:

            if type(test) is not dict:
                logger.error('test not dict type, return None')
                return None

            context = None
            if 'context' in test.keys():
                context = test['context']

            if 'matches' not in test:
                logger.error('test missing "matches" attribute, return None')
                return None

            if type(test['matches']) is list:
                expressions = test['matches']
            else:
                expressions = [str(test['matches'])]

            return self.__test_in_context(expressions, context, data)

    def __test_in_context(self, expressions, context, data):

        if type(data) is dict and context is None:
            for item_k, item_v in data.items():
                string = str(item_v)
                expressions_match_count = 0
                for expression in expressions:
                    if re.search(expression, string):
                        logger.debug('within context {} matched {} in {}'.format(item_k, expression, string))
                        expressions_match_count += 1
                if expressions_match_count == len(expressions):
                    return {
                        'matches': expressions,
                        'context': string[0:120] + ' ...'
                    }

        elif type(data) is dict and context in data:
            string = str(data[context])
            expressions_match_count = 0
            for expression in expressions:
                if re.search(expression, string):
                    logger.debug('within context {} matched {} in "{}"'.format(context, expression, string))
                    expressions_match_count += 1
            if expressions_match_count == len(expressions):
                return {
                    'matches': expressions,
                    'context': string[0:120] + ' ...'
                }

        elif type(data) is dict and context not in data:
            logger.warn('context "{}" not found in data'.format(str(context)))

        else:
            logger.error('unsupported data type presented for {}'.format(str(expressions)))

        return None




import whois
from . import PhishingTrackerMeta


class PhishingTrackerWhois:

    @staticmethod
    def analyzer(reference):
        try:
            data = dict(whois.whois(PhishingTrackerMeta.domain_name(reference)))
        except Exception as e:
            data = {'exception': str(e)}
        return data




from smtplib import SMTP


class PhishingTrackerSmtp:

    @staticmethod
    def analyzer(mx_records):

        if type(mx_records) is not list or len(mx_records) == 0:
            return None

        data = {}

        for mx_record in mx_records:
            if ' ' not in mx_record:
                continue
            _, mx_record_hostname = mx_record.split(' ')

            smtp_response = SMTP().connect(mx_record_hostname)
            if type(smtp_response) is tuple:
                smtp_code, smtp_header = smtp_response
                data[mx_record_hostname] = {
                    'status_code': smtp_code,
                    'header': smtp_header.decode('utf8')
                }

        if len(data) == 0:
            return None

        return data

    @staticmethod
    def test(tests, data):
        pass




import tldextract
from urllib.parse import unquote_plus


class PhishingTrackerMeta:

    @staticmethod
    def meta_data(reference):
        return {
            'reference': reference,
            'host_name': PhishingTrackerMeta.host_name(reference),
            'domain_name': PhishingTrackerMeta.domain_name(reference),
            'url_decode': PhishingTrackerMeta.url_decode(reference),
        }

    @staticmethod
    def domain_name(reference):
        return unquote_plus(tldextract.extract(reference).registered_domain.lower())

    @staticmethod
    def host_name(reference):
        return unquote_plus(tldextract.extract(reference).fqdn.lower())

    @staticmethod
    def url_decode(reference, prefix='https'):

        #  www.foobar.com
        #  www.foobar.com/foo.bar.html
        if not reference.startswith('http'):
            reference = prefix + '://' + reference

        #  //www.foobar.com/foo/bar.html
        if reference.startswith('//'):
            reference = prefix + ':' + reference

        return unquote_plus(reference)




import requests
from .useragent import user_agent

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class PhishingTrackerWeb:

    @staticmethod
    def analyzer(url):

        headers = {
            'User-Agent': user_agent(),
            'Cache-Control': 'no-cache'
        }

        try:
            r = requests.get(url, headers=headers, verify=False, allow_redirects=False)
        except Exception as e:
            return {'exception': str(e)}

        data = {
            'request': {
                'url': url,
                'headers': dict(r.request.headers),
            },
            'response': {
                'status_code': r.status_code,
                'headers': dict(r.headers),
                'text': r.text,
            }
        }

        return data

    @staticmethod
    def test(tests, data):
        pass




import os
import ssl
import socket
import tempfile

from .util import json_dumps
from .util import datetime_parse


class PhishingTrackerCertificate:

    @staticmethod
    def analyzer(hostname, port=443, append_raw_certificate=True):
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        sock = context.wrap_socket(conn, server_hostname=hostname)

        try:
            sock.connect((hostname, port))
        except Exception as e:
            return 'Exception: {}'.format(str(e))

        certificate_raw = ssl.DER_cert_to_PEM_cert(sock.getpeercert(True))
        certificate = PhishingTrackerCertificate.__decode_raw_certificate(certificate_raw)

        if append_raw_certificate is True:
            certificate['raw'] = certificate_raw
        return certificate

    @staticmethod
    def __decode_raw_certificate(certificate_raw):

        filename = tempfile.mktemp()

        with open(filename, 'w') as f:
            f.write(certificate_raw)

        try:
            # using and abusing a private _ssl function
            data = dict(ssl._ssl._test_decode_cert(filename))
            os.unlink(filename)
        except Exception:
            return None

        if data is None or len(data) == 0:
            return None

        def _rollup_a(input):
            if type(input) is not tuple:
                return None
            __data = {}
            for item in input:
                attrib, value = item[0]
                __data[attrib] = value
            return __data

        def _rollup_b(input):
            if type(input) is not tuple:
                return None
            __data = {}
            for item in input:
                attrib, value = item
                if attrib in __data:
                    __data[attrib].append(value)
                else:
                    __data[attrib] = [value]
            return __data

        if 'issuer' in data:
            data['issuer'] = _rollup_a(data['issuer'])

        if 'subject' in data:
            data['subject'] = _rollup_a(data['subject'])

        if 'subjectAltName' in data:
            data['subjectAltName'] = _rollup_b(data['subjectAltName'])

        if 'notBefore' in data:
            data['notBefore'] = datetime_parse(data['notBefore'])

        if 'notAfter' in data:
            data['notAfter'] = datetime_parse(data['notAfter'])

        return data

    @staticmethod
    def test(tests, data):
        pass




import os
import tempfile
import pytest
from unittest.mock import patch
import PhishingTracker


def test_name_exist():
    pt = PhishingTracker
    assert pt.NAME is not None


def test_version_exist():
    pt = PhishingTracker
    assert pt.VERSION is not None



#!/usr/bin/env python3

from setuptools import setup, find_packages
from PhishingTracker import NAME
from PhishingTracker import VERSION

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=VERSION,
    description='Utility to manage sets of phishing links making it easier to track their removal progress over time.',

    long_description=long_description,
    long_description_content_type='text/markdown',

    classifiers=[
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: BSD License',
    ],
    keywords=['phish', 'phishing', 'url', 'cyber-crime'],

    author='Nicholas de Jong',
    author_email='contact@nicholasdejong.com',
    url='https://github.com/ndejong/phishing-tracker',
    license='BSD 2-Clause',

    packages=find_packages(),
    zip_safe=False,
    scripts=['bin/phishing-tracker'],

    install_requires=[
        'pydig',
        'pyyaml',
        'requests',
        'dateparser',
        'tldextract',
        'python-whois',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],

)




import os
import yaml
import hashlib

from .util import json_dumps


class PhishingTrackerFile:

    def __init__(self, __logger=None):
        global logger
        logger = __logger

    def load_config(self, filename=None):

        config = {}
        try:
            with open(filename, 'r') as f:
                config = yaml.safe_load(f.read())
        except Exception as e:
            logger.fatal(e)
            exit(1)
        logger.info('Loaded {} items from {}'.format(len(config), filename))
        return config

    def save_datafile(self, data, timestamp, pathname=None):

        subpath = os.path.join(data['meta']['domain_name'])
        if pathname is not None:
            subpath = os.path.join(pathname, data['meta']['domain_name'])

        if not os.path.isdir(subpath):
            os.mkdir(subpath)

        md5prefix = hashlib.md5(data['meta']['reference'].encode('utf-8')).hexdigest()[0:4]

        filename = os.path.join(subpath, '{}_{}_{}.json'.format(data['meta']['host_name'], md5prefix, timestamp))
        with open(filename, 'w') as f:
            f.write(json_dumps(data, indent='  '))

        logger.info('Phish data written to {}'.format(filename))




import pydig


class PhishingTrackerDig:

    @staticmethod
    def analyzer(record, type='ANY'):

        type = type.upper()

        if type == 'ANY':
            types = [ 'A', 'AAAA', 'CNAME', 'MX', 'NS', 'TXT' ]
            data = {}
            for t in types:
                data[t] = PhishingTrackerDig.analyzer(record, t)
            return data

        try:
            response = pydig.query(record, type)
        except Exception as e:
            response = 'Exception: {}'.format(str(e))

        if type == 'TXT':
            response = [s.strip('"') for s in response]
        return response

    @staticmethod
    def test(tests, data):
        pass




from . import NAME
from . import VERSION
from . import analyzers_available

from .util import json_dumps
from .util import timestamp

from . import PhishingTrackerLogger
from . import PhishingTrackerFile
from . import PhishingTrackerMeta
from . import PhishingTrackerAnalyzers

from . import PhishingTrackerTests


class PhishingTracker:

    debug = None

    def __init__(self, debug=False):

        self.debug = debug

        log_level = 'info'
        if self.debug:
            log_level = 'debug'

        global logger
        logger = PhishingTrackerLogger(level=log_level).logger
        logger.info('{} v{}'.format(NAME, VERSION))

    def main(self, reference=None, filename=None, pathname=None, analyzers=None):

        ts = timestamp()

        if reference is not None:
            self.playbook(reference, analyzers, None, pathname, ts)

        elif filename is not None:
            for item in PhishingTrackerFile(logger).load_config(filename=filename).items():
                if type(item) is not tuple:
                    logger.error('Item found in file is unsupported format: {}'.format(str(item)))
                else:
                    item_analyzers = analyzers
                    item_reference, item_config = item
                    if item_config is None or 'enabled' not in item_config.keys() or \
                            ('enabled' in item_config and item_config['enabled']):
                        self.playbook(item_reference, item_analyzers, item_config, pathname, ts)
        return

    def playbook(self, reference, analyzers, config, pathname, ts):

        # meta
        data = { 'meta': PhishingTrackerMeta.meta_data(reference=reference) }

        logger.info('reference: {}'.format(data['meta']['reference']))
        logger.debug('host_name: {}'.format(data['meta']['host_name']))
        logger.debug('domain_name: {}'.format(data['meta']['domain_name']))
        logger.debug('url_decode: {}'.format(data['meta']['url_decode']))

        # analyzers setup
        if analyzers is None or (type(analyzers) is list and analyzers[0] is None):
            if type(config) is dict and 'analyzers' in config.keys():
                analyzers = config['analyzers']
            else:
                analyzers = analyzers_available
        if 'all' in analyzers:
            analyzers = analyzers_available

        # analyzers
        data['analyzers'] = PhishingTrackerAnalyzers(logger).analyzers(data=data, analyzers=analyzers)

        # analyzer_report_sets_expected setup
        if config is not None and 'analyzer_report_sets_expected' in config:
            analyzer_report_sets_expected = config['analyzer_report_sets_expected']
        else:
            analyzer_report_sets_expected = None

        # analyzer_report_sets_not_expected setup
        if config is not None and 'analyzer_report_sets_not_expected' in config:
            analyzer_report_sets_not_expected = config['analyzer_report_sets_not_expected']
        else:
            analyzer_report_sets_not_expected = None

        # analyzers_report
        data['analyzers_report'] = PhishingTrackerAnalyzers(logger).analyzers_report(
            data=data,
            analyzer_report_sets_expected=analyzer_report_sets_expected,
            analyzer_report_sets_not_expected=analyzer_report_sets_not_expected,
        )

        # analyzer_report_sets - logging
        if 'analyzer_report_sets' in data['analyzers_report'] and data['analyzers_report']['analyzer_report_sets']:
            if 'expected' in data['analyzers_report']['analyzer_report_sets'] and data['analyzers_report']['analyzer_report_sets']['expected'] is False:
                logger.warn('analyzer_report_sets_expected returns false!')
            if 'not_expected' in data['analyzers_report']['analyzer_report_sets'] and data['analyzers_report']['analyzer_report_sets']['not_expected'] is True:
                logger.warn('analyzer_report_sets_not_expected returns true!')

        # tests
        if type(config) is dict and 'tests' in config.keys() and len(config['tests']) > 0:
            data['tests_report'] = PhishingTrackerTests(logger).tests(data=data, tests=config['tests'])

        PhishingTrackerFile(logger).save_datafile(data, ts, pathname)
        self.report_output(data, analyzers_report=True, tests_report=True)

    def report_output(self, data, analyzers_report=True, tests_report=True):
        if analyzers_report is True and 'analyzers_report' in data:
            print(json_dumps({'analyzers_report': data['analyzers_report']}, indent='  '))
        if tests_report is True and 'tests_report' in data and len(data['tests_report']) > 0:
            print(json_dumps({'tests_report': data['tests_report']}, indent='  '))
